# -*- coding: utf-8 -*-

import pytest
from django.utils.timezone import now

from chrono.agendas.models import Agenda, Booking, Category, Event

pytestmark = pytest.mark.django_db


def test_statistics_list(app, user):
    Category.objects.create(label='Category A')
    Category.objects.create(label='Category B')

    # unauthorized
    app.get('/api/statistics/', status=401)

    app.authorization = ('Basic', ('john.doe', 'password'))
    resp = app.get('/api/statistics/')
    category_filter = [x for x in resp.json['data'][0]['filters'] if x['id'] == 'category'][0]
    assert len(category_filter['options']) == 3


def test_statistics_bookings(app, user, freezer):
    agenda = Agenda.objects.create(label='Foo bar', kind='events')
    event = Event.objects.create(start_datetime=now(), places=5, agenda=agenda)

    app.authorization = ('Basic', ('john.doe', 'password'))
    resp = app.get('/api/statistics/')
    url = [x for x in resp.json['data'] if x['id'] == 'bookings_count'][0]['url']

    resp = app.get(url)
    assert len(resp.json['data']['series']) == 0

    freezer.move_to('2020-10-10')
    for _ in range(10):
        Booking.objects.create(event=event)
    freezer.move_to('2020-10-15')
    Booking.objects.create(event=event)

    resp = app.get(url + '?time_interval=day')
    assert resp.json['data'] == {
        'x_labels': ['2020-10-10', '2020-10-15'],
        'series': [{'label': 'Bookings Count', 'data': [10, 1]}],
    }

    # period filter
    resp = app.get(url + '?start=2020-10-14&end=2020-10-16')
    assert resp.json['data'] == {
        'x_labels': ['2020-10-15'],
        'series': [{'label': 'Bookings Count', 'data': [1]}],
    }

    category = Category.objects.create(label='Category A', slug='category-a')
    agenda = Agenda.objects.create(label='Foo bar', kind='events', category=category)
    event = Event.objects.create(start_datetime=now(), places=5, agenda=agenda)
    freezer.move_to('2020-10-25')
    Booking.objects.create(event=event)

    # category filter
    resp = app.get(url + '?category=category-a')
    assert resp.json['data'] == {
        'x_labels': ['2020-10-25'],
        'series': [{'label': 'Bookings Count', 'data': [1]}],
    }

    # invalid time_interval
    resp = app.get(url + '?time_interval=month', status=400)
    assert resp.json['err'] == 1
    assert 'time_interval' in resp.json['errors']

    # absence/presence
    for i in range(10):
        Booking.objects.create(event=event, user_was_present=bool(i % 2))

    freezer.move_to('2020-11-01')
    Booking.objects.create(event=event, user_was_present=True)

    resp = app.get(url)
    assert resp.json['data'] == {
        'x_labels': ['2020-10-10', '2020-10-15', '2020-10-25', '2020-11-01'],
        'series': [
            {'label': 'Unknown', 'data': [10, 1, 1, None]},
            {'label': 'Present', 'data': [None, None, 5, 1]},
            {'label': 'Absent', 'data': [None, None, 5, None]},
        ],
    }
